
import sympy, numbers

from .taylornum import taylornum
from .taylorsym import taylorsym


# Calculate the taylor series.
# This is the main routing function that determines symbolic or numerical Taylor series
# and where to send the data to
def taylor(f, x0, n=None):
    # If n is None, then we know that this is a numeric
    if n is None:
        return taylornum(f, x0)

    # Symbolic. Check that everytfalsehing is a symbol
    if isinstance(f, list) or isinstance(f, tuple):
            if all(isinstance(fi, tuple(sympy.core.all_classes)) or isinstance(fi, numbers.Number) for fi in f):
                if isinstance(x0, list) or isinstance(x0, tuple):
                    if len(f) == len(x0):
                        return taylorsym(f, x0, n)
                    else:
                        raise RuntimeError('ERROR: Size of f and x0 do not match')
                else:
                    raise RuntimeError('x0 is not the same size as f')
            else:
                raise RuntimeError('ERROR: Expected the list to be all symbolic expressions.')
    elif isinstance(f, tuple(sympy.core.all_classes)) or isinstance(f, numbers.Number):
        return taylorsym(f, x0, n)
    else:
        raise RuntimeError('ERROR: Expected symbolic expressions.')



import sympy, numbers

from ohtaylor.taylorutils import ktuples_nsum, factorial_tuple, create_symbol_point_list
from ohtaylor.diffcalc import *


# Find the Taylor series when f:R^1 -> R^1
def taylorsym_1_1(x, f, x0, n):
    # If the input is given as a list or tuple of one variable, change it to be a number
    if isinstance(x0, list) or isinstance(x0, tuple):
        x0 = x0[0]

    px = f.subs(x, x0)
    fi_1x = f

    for i in range(1,n+1):
        fix = sympy.diff(fi_1x, x)
        fi_1x = fix
        px += fix.subs(x, x0) / sympy.factorial(i) * (x - x0)**i
    return px


# Find the Taylor series for f:R^N -> R
def taylorsym_N_1(x, f, x0, n):
    # Number of input variables
    k = len(x)
    vars = create_symbol_point_list(x, x0)
    d = DiffCalcCached()
    syms, ivs = zip(*vars)

    # Generate the Taylor series
    px = 0
    for i in range(0, n+1):
        tuples = ktuples_nsum(k, i)
        for t in tuples:
            fix = d.diff(f, syms, t)
            fix = fix.subs(vars)
            for j in range(0, k):
                fix *= (syms[j] - ivs[j])**t[j]
            px += fix / factorial_tuple(t)
    return px


# Send to R^1 or R^n
def taylorsym_d(f, x0, n):
    if isinstance(f, numbers.Number):
        return f

    # Get the independent variables
    symbols = f.free_symbols

    # Dispatch to the right place
    if len(symbols) == 1:
        [x] = symbols
        return taylorsym_1_1(x, f, x0, n)
    else:
        return taylorsym_N_1(symbols, f, x0, n)


# Send each of the dimensions of R^M output to be evaluated independently
def taylorsym_l(f, x0, n):
    px = [None]*len(f)
    for i in range(0, len(f)):
        px[i] = taylorsym_d(f[i], x0[i], n)
    return px


# Figure out the dimensions and dispatch to the right Taylor series evaluation
# Check that the inputs are valid and match up
def taylorsym(f, x0, n):
    if isinstance(f, list) or isinstance(f, tuple):
        return taylorsym_l(f, x0, n)
    else:
        return taylorsym_d(f, x0, n)





import unittest
from sympy.abc import x, y, z

from ohtaylor import *
from ohtaylor.taylornum import *
from ohtaylor.taylorutils import *
from ohtaylor.diffcalc import *

# return value has a string if false and just true when
def get_flag(val):
    if isinstance(val, tuple):
        return val[0]
    else:
        return val

# Test the numerical Taylor series calculation
class TestNumericalTaylor(unittest.TestCase):

    # Test the single variable exponential function
    def test_exp_single_var(self):
        # Linear approximation of e^x at 0.
        # Should be e^0 + e^0*dt
        f = [1, 1]
        dt = 0.1
        self.assertAlmostEqual(1.1, taylor(f, dt))

        # Quadratic approximation of e^x at 0
        # Should be e^0 + e^0*dt + e^0*dt^2
        f = [1, 1, 1]
        dt = 0.1
        self.assertAlmostEqual(1.105, taylor(f, dt))

    # Test the single variable sine Taylor series expansion
    def test_sine_single_var(self):
        # Linear approximation of sin(x) at 0
        # Should be sin(0) + cos(0)*dt
        f = [0, 1]
        dt = 0.1
        self.assertAlmostEqual(0.1, taylor(f, dt))

        # Quadratic approximation of sin(x) at 0
        # Should be sin(0) + cos(0)*dt + sin(0)*dt^2
        f = [0, 1,0]
        dt = 0.1
        self.assertAlmostEqual(0.1, taylor(f, dt))

    # Test when we have multiple R->R^M. The input data for f is a list, then we
    # know it has multi-dimensional output
    def test_R_to_R2(self):
        f1 = (1, 1, 1)
        dt1 = 0.1
        f2 = (0, 1)
        dt2 = 0.1
        solution = (1.105, 0.1)
        self.assertAlmostEqual(solution[0], taylor((f1,f2), (dt1, dt2))[0])
        self.assertAlmostEqual(solution[1], taylor((f1, f2), (dt1, dt2))[1])

    # Check that the checker for the dictionary input works and that it fails on bad input
    def test_rn_r_dict_check(self):
        # Check dictionary without tuple
        fn = {}
        fn[0] = 1
        self.assertFalse(check_taylornum_RN_R(fn)[0])

        # Check not all are tuples
        fn.clear()
        fn[(0,)] = 1
        fn[0] = 1
        self.assertFalse(check_taylornum_RN_R(fn)[0])

        # Check different size tuples
        fn.clear()
        fn[(0,)] = 1
        fn[(0,1)] = 1
        self.assertFalse(check_taylornum_RN_R(fn)[0])

        # Check that all the partial derivatives are present
        fn.clear()
        fn[(0,0,0)] = 1
        fn[(0,1,0)] = 1
        fn[(0,0,1)] = 1
        self.assertFalse(check_taylornum_RN_R(fn)[0])

        # Check that we are missing some of the partial derivatives
        fn.clear()
        fn[(0,)] = 1
        fn[(2,)] = 1
        self.assertFalse(check_taylornum_RN_R(fn)[0])

        # Check a good one that is R->R
        fn.clear()
        fn[(0,)] = 1
        fn[(1,)] = 1
        fn[(2,)] = 1
        self.assertTrue(check_taylornum_RN_R(fn)[0])

    # When the first variable is a dictionary, then we assume that all the derivatives
    # are given as tuples in the dictionary
    def test_rn_tor(self):
        # Simple single variable case
        fn = {}
        fn[(0,)] = 1
        fn[(1,)] = 1
        fn[(2,)] = 1
        dt = (0.1,)

        val = taylor(fn, dt)
        self.assertAlmostEqual(1.105, val)

        # This is from the function exp(x+y) and we want it to close to what exp gives
        fn.clear()
        fn[(0, 0)] = 1
        fn[(1, 0)] = 1
        fn[(0, 1)] = 1
        fn[(2, 0)] = 1
        fn[(1, 1)] = 1
        fn[(0, 2)] = 1
        dt = (0.05, 0.05)
        val = taylor(fn, dt)
        self.assertAlmostEqual(1.105, val)
        dt = (0.01, 0.09)
        self.assertAlmostEqual(1.105, taylor(fn, dt))

    # Sending symbols rather than dt.
    # Technically (x - x0) = (x0 + dt - x0) = dt
    def test_dt_as_sym(self):
        # Simple single variable case
        fn = {}
        fn[(0,)] = 1
        fn[(1,)] = 1
        fn[(2,)] = 1
        dt = (0.1,)

        val = taylor(fn, (x,))
        px = taylor(sympy.exp(x), n=2, x0=0)
        self.assertEqual(val, px)

    # Function to compute all the derivatives of a function at a given point and given n
    # Return a dictionary of the values of the derivatives for the corresponding tuples
    @staticmethod
    def compute_derivative(f, n, x0):
        # Dictionary where we will store the values
        fvals = {}
        # Derivative calculator
        d = DiffCalcCached()

        # Unzip the values
        syms, _ = zip(*x0)

        for i in range(0, n+1):
            # Create the list for tuples for this derivative
            tuples = ktuples_nsum(len(syms), i)
            for t in tuples:
                # Calculate the derivatives and add it to the dictionary
                fi = d.diff(f, syms, t)
                fix = (fi.subs(x0)).evalf()
                fvals[t] = fix
        return fvals

    # Test that the compute derivatives symbolically to compute the dictionary
    # works correctly.
    def test_compute_derivatives(self):
        # Check that all the derivatives of exp(x) are 1
        f = sympy.exp(x)
        d = TestNumericalTaylor.compute_derivative(f, 3, tuple(zip((x,), (0,))))
        self.assertEqual(4, len(d))
        self.assertAlmostEqual(1.0, d[(0,)])
        self.assertAlmostEqual(1.0, d[(1,)])
        self.assertAlmostEqual(1.0, d[(2,)])
        self.assertAlmostEqual(1.0, d[(3,)])

        # Check that derivatives of higher dimensions are also calculated correctly
        f = sympy.exp(x*y)
        d = TestNumericalTaylor.compute_derivative(f, 2, tuple(zip((x,y), (0,0))))
        self.assertEqual(6, len(d))
        self.assertAlmostEqual(1.0, d[(1,1)])

    # Make sure that the sym and num produce the same results for the given polynomials
    def test_num_with_sym(self):
        # Function that we will like to test out to
        f1 = sympy.functions.elementary.exponential.exp(x + y)
        f2 = sympy.functions.elementary.trigonometric.cos(x ** 2 + y + z)
        f = (f1, f2)
        x0 = (0, 0)
        x1 = (0, 0, 1)
        p = taylor(f, (x0, x1), 2)

        # Create the numerical values for the partial derivatives from what was calculated
        # and check if they match up
        n = 6
        syms = (x, y)
        fvals1 = TestNumericalTaylor.compute_derivative(f1, n, tuple(zip(syms, x0)))
        dt1 = (0.1, 0.1)
        pxn1 = taylor(fvals1, dt1)
        self.assertAlmostEqual(f1.subs(tuple(zip(syms, tuple(map(lambda x, y: x+y, x0, dt1))))), pxn1)

        # Check the two dimensional values
        n = 7
        syms = (x, y, z)
        fvals2 = TestNumericalTaylor.compute_derivative(f2, n, tuple(zip(syms, x1)))

        # Use the calculated values to do the Taylor series numerical approximation
        dt2 = (0.1, 0.1, 0.1)
        x1dt = tuple(map(lambda x, y : x+y, x1, dt2))
        px = f2.subs(tuple(zip(syms, x1dt))).evalf()
        pxn2 = taylor(fvals2, dt2)
        self.assertAlmostEqual(px, pxn2)

        # f is R^3->R^2
        f = (fvals1, fvals2)
        px = taylor(f, (dt1, dt2))
        self.assertAlmostEqual(pxn1, px[0])



import unittest, math, sympy
from sympy.abc import x, y, z

from ohtaylor.taylorutils import *


# Test the utility function that calculates the tuples of partitions
class TestTaylorUtils(unittest.TestCase):

    # Test that tuples sum works the way we want it to
    def test_tuples_sum(self):
        data = list(ktuples_nsum(0, 0))
        self.assertEqual((), data[0])

        # Singleton should only create a single value
        data = list(ktuples_nsum(1, 1))
        self.assertEqual(1, len(data))
        self.assertEqual((1,), data[0])
        self.assertEqual(1, nkparts(1, 1))

        # Check that all the permutations are generated
        data = list(ktuples_nsum(2, 2))
        self.assertEqual(3, len(data))
        self.assertTrue((2,0) in data)
        self.assertTrue((0, 2) in data)
        self.assertTrue((1, 1) in data)
        self.assertEqual(3, nkparts(2, 2))

        self.assertEqual(2, nkparts(2, 1))
        data = list(ktuples_nsum(3, 2))
        self.assertEqual(len(data), nkparts(3, 2))

        data = list(ktuples_nsum(4, 12))
        self.assertTrue((3, 3, 3, 3) in data)
        self.assertTrue((0, 0, 12, 0) in data)
        self.assertEqual(len(data), nkparts(4, 12))
        # Check that there are no duplicates
        self.assertEqual(len(data), len(set(data)))

    # Test taking the factorial of a tuple
    def test_factorial_tuple(self):
        t = (1, 2, 3)
        v = math.factorial(1) * math.factorial(2) * math.factorial(3)
        self.assertEqual(v, factorial_tuple(t))

        t = (3, 2, 1, 5, 7)
        v = math.factorial(3) * math.factorial(2) * math.factorial(5) * math.factorial(7)
        self.assertEqual(v, factorial_tuple(t))

    # Sort symbol list and check we have symbols that are sorted by their string representation
    def test_symbols_sort(self):
        f = z + y + x
        l = list(f.free_symbols)
        l = sort_symbols(l)
        self.assertEqual([x, y, z], l)
        self.assertEqual([x, y, z], sort_symbols([z, y, x]))

        pi = sympy.symbols('pi')
        theta = sympy.symbols('theta')
        kappa = sympy.symbols('kappa')
        l = sort_symbols([kappa, theta, pi])
        self.assertEqual([kappa, pi, theta], l)

    # Check that the symbol-point list is created correctly
    def test_create_symbol_point_list(self):
        f = z + y + x
        x0 = [1, 2, 3]
        self.assertEqual([(x, 1), (y, 2), (z, 3)], create_symbol_point_list(f.free_symbols, x0))

        l1 = [(z, 3), (y, 2), (x, 1)]
        self.assertEqual(l1, create_symbol_point_list(f.free_symbols, l1))




import sympy

# Generator of tuples of size k that sum up to n using recursion
def ktuples_nsum(k, n):
    if k == 0 and n == 0:
        yield tuple()
        return
    if k == 1:
        yield (n,)
        return
    if n == 0:
        yield (0,)*k
        return

    for start in range(n, -1, -1):
        for qu in ktuples_nsum(k-1, n-start):
            yield (start,)+qu


# Calculate the number of tuples of size k that sum to n (number of tuples the above function calculates)
def nkparts(k, n):
    if k == 1:
        return 1
    if k == 2:
        return n + 1

    sum = 0
    for i in range(0, n+1):
        sum += nkparts(k-1, n-i)
    return sum


# For each of the integer in the tuple, calculate the factorial and multiply them together
def factorial_tuple(t):
    if isinstance(t, tuple):
        v = 1
        for i in range(0, len(t)):
            v *= sympy.factorial(t[i])
        return v
    else:
        raise RuntimeError("Input to the function must be a tuple")


# Sort the symbols by their
def sort_symbols(symbols):
    return sorted(symbols, key=lambda k: k.name)


# Create a list of tuples of symbols and values. If the input is already in that format,
# do nothing
def create_symbol_point_list(symbols, x0):
    # Check that the dimensions of symbols and x0 match up
    if (len(symbols) != len(x0)):
        raise RuntimeError("Dimension of x_0 does not match the number of symbols. Symbols %d, x0: %d" % (len(symbols),  len(x0)))

    # Check if x0 is already in the right format
    if all(type(x) is tuple and len(x) == 2 for x in x0):
        return x0

    vals = sort_symbols(symbols)
    return list(zip(vals, x0))





import math

from ohtaylor.taylorutils import nkparts, ktuples_nsum, factorial_tuple


def taylornum_R1_R1(fvals, dt):
    px = 0
    for i in range(0, len(fvals)):
        px += fvals[i] / math.factorial(i) * math.pow(dt, i)
    return px


# Check that the input dictionary in R^N->R is valid
def check_taylornum_RN_R(fvals):
    # Check that all the entries are tuples to floats
    if not all (isinstance(key, tuple) for key in fvals.keys()):
        return False, 'Entries not tuples'

    # Estimate n from the first tuple entry and check that all the tuples
    # have that length
    N = len(list(fvals.keys())[0])
    if not all (len(key) == N for key in fvals.keys()):
        return False, 'Tuple lengths not the same'

    # Check that the size of the dictionary is correct and that the right
    # number of partials are present
    K = 0
    sumk = 0
    # First get the length of the dictionary
    nentries = len(fvals)
    while sumk < nentries:
        sumk += nkparts(N, K)
        K += 1
    if sumk != nentries:
        return False, 'Number of entries are not consistent'

    # Check that all the partial derivatives are present for the given n
    for k in range(0, K):
        dtuples = ktuples_nsum(N, k)
        if not all( dtuple in fvals for dtuple in dtuples):
            return False, 'Missing derivative'

    return True, (N, K)


# Handle the case of going to one output dimension
def taylornum_RN_R(fvals, dt):
    # Check the we have good values in the dictionary
    f,v = check_taylornum_RN_R(fvals)
    if not f:
        raise RuntimeError(v)

    # Create the polynomial to evaluate the Taylor series
    N, K = v
    fdt = 0
    for k in range(0, K):
        tuples = ktuples_nsum(N, k)
        for t in tuples:
            fix = fvals[t]/factorial_tuple(t)
            for j in range(0, N):
                fix *= dt[j] ** t[j]
            fdt += fix

    return fdt


# Handle the case of RM multiple outputs
def taylornum_RN_RM(fvals, dt):
    if len(fvals) != len(dt):
        raise RuntimeError("Dimensions of f and dt do not agree")

    pdt = []
    for i in range(0, len(fvals)):
        if isinstance(fvals[i], dict):
            pdt.append(taylornum_RN_R(fvals[i], dt[i]))
        else:
            pdt.append(taylornum_R1_R1(fvals[i], dt[i]))
    return pdt


def taylornum(fvals, dt):
    if isinstance(fvals, dict):
        return taylornum_RN_R(fvals, dt)

    # List or tuple means that output is R^M. Dict means that input is also R^N
    if all (isinstance(fivals, (list, tuple, dict)) for fivals in fvals ):
        return taylornum_RN_RM(fvals, dt)

    # This is a R->R function
    [pdt] = taylornum_RN_RM((fvals,), (dt,))
    return pdt





import unittest, sympy, math
from sympy.abc import x, y, z

from ohtaylor import *


# Check the symbolic evaluation of Taylor series
class TestSymbolicTaylor(unittest.TestCase):

    # Test the calculation of Taylor series for e^x
    def test_exp_single_var(self):
        # Given exp(x), the Taylor series for n=1 at x0=0 is 1 + 1(1-dx)
        x = sympy.symbols('x')
        f = sympy.functions.elementary.exponential.exp(x)
        x0 = 0
        px = taylor(f, x0, 1)
        #sympy.pprint(px)
        self.assertTrue(px == x + 1)
        # Evaluate the function at some points and check values are correct
        self.assertAlmostEqual(1.1, px.evalf(subs={x: 0.1}))

        # When n=2, we have P(x) = 1 + x + x^2/2
        px = taylor(f, x0, 2)
        #sympy.pprint(px)
        self.assertTrue(px == 1 + x + x**2/2)
        # Evaluate the function at a point
        self.assertAlmostEqual(1.105, px.evalf(subs={x: 0.1}))

        # Evaluate the value of exp(x) at 1 to calculate the value of e
        px = taylor(f, x0, 20)
        self.assertAlmostEqual(2.718281828459045, px.evalf(subs={x: 1}))

    # Test tje calculation of Taylor series for sin(x)
    def test_sine_single_var(self):
        # Evaluate the Taylor series of sin(x)
        f = sympy.functions.elementary.trigonometric.sin(x)
        x0 = 0
        px = taylor(f, x0, 1)
        #sympy.pprint(px)
        self.assertTrue(px == x)
        px = taylor(f, x0, 2)
        self.assertTrue(px == x)
        px = taylor(f, x0, 3)
        self.assertTrue(px == x - x**3/6)

        # Evaluate sin(x) at 45 degrees with x0 at 0 and 1
        self.assertAlmostEqual(math.sin(math.pi/4), taylor(f, 0, 10).evalf(subs={x: math.pi/4}))
        self.assertAlmostEqual(math.sin(math.pi / 4), taylor(f, 1, 10).evalf(subs={x: math.pi / 4}))

    # Test the calculation of Taylor series for [e^x e^x] at [0,1]
    def test_exp_1toM(self):
        # Evaluate the Taylor series of [e^x e^x] which is just e^x repeated twice in a list
        f = [sympy.functions.elementary.exponential.exp(x), sympy.functions.elementary.exponential.exp(x)]
        x0 = [0, 1]
        px = taylor(f, x0, 2)
        #sympy.pprint(px)
        self.assertTrue(px[0] == 1 + x + x ** 2 / 2)
        e_ = sympy.functions.elementary.exponential.exp(1)
        self.assertTrue(px[1] == e_ + e_*(x-1) + e_*(x - 1) ** 2 / 2)

    # Check the error handling of the taylor function when the input is bad and when
    # we mix expressions with numbers
    def test_bad_input(self):
        # Check that a number is just returned back
        self.assertEqual(1, taylor(1, 2, 3))
        self.assertEqual([1, 2.1], taylor([1, 2.1], [1,2], 3))
        with self.assertRaises(RuntimeError) as context:
            taylor([1, 2], 2, 3)
        with self.assertRaises(RuntimeError) as context:
            taylor([1,x], 2, 3)
        with self.assertRaises(RuntimeError) as context:
            taylor([[x], [x]], 2, 3)
        self.assertEqual([1, taylor(x**2, 1, 3)], taylor([1, sympy.symbols('x')**2], [2,1], 3))

    # Test the calculation of Taylor series for exp(xy)
    def test_exp_multi_var(self):
        f = sympy.functions.elementary.exponential.exp(x**2 + y)
        x0 = [0, 0]
        px = taylor(f, x0, 3)
        #sympy.pprint(px)
        self.assertEqual(1 + y + x**2 + y**2/2 + x**2*y + y**3/6, px)

        # Check if this still works when we evaluate at exp((1,1)) and match to exp(2)
        px = taylor(f, x0, 10)
        self.assertAlmostEqual(math.exp(.2), px.subs([(x,math.sqrt(.1)), (y,.1)]).evalf())

    def test_sin_multi_var(self):
        f = sympy.functions.elementary.trigonometric.sin(x + y)
        x0 = [0, 0]
        px = taylor(f, x0, 10)
        #sympy.pprint(px)

        # Evaluate the value when x + y = 45 degrees (pi/4) and then see if the value matches
        self.assertAlmostEqual(math.sin(math.pi/4), px.subs([(x, math.pi/8), (y, math.pi/8)]).evalf())
        self.assertAlmostEqual(math.sin(math.pi / 4), px.subs([(x, 3*math.pi / 4), (y, -math.pi / 2)]).evalf())
    # Make a multi-variable sin(x+y) and check that output value is correct when we plug in values

    # Check that calculating the Taylor series of a polynomial is a polynomial itself
    def test_poly_multi_var(self):
        f = x**3 + 2*y - y**3 - 3*x
        x0 = [0, 0]
        px = taylor(f, x0, 10)
        self.assertTrue(px == f)

    # Function with three input variables and one output variable
    def test_multi_var_with_3_inputs(self):
        f = x ** 3 * y * z
        x0 = [1, 3, -2]
        px = taylor(f, x0, 5)
        xval = [(x, 1/2), (y, 2/3), (z, 1/2)]
        self.assertAlmostEqual(f.subs(xval), px.subs(xval))

    # Example from http://www.math.ubc.ca/~feldman/m226/taylor2d.pdf
    def test_sqrt_multi_var(self):
        f = sympy.sqrt(1 + y**2 + 4*x**2)
        x0 = [1,2]
        # Send the values as a list of tuples so that we know which part is x and y
        px = taylor(f, list(zip([x,y], x0)), 2)
        #sympy.pprint(sympy.expand(px))
        epx = 3 + sympy.Rational(4,3)*(x-1) + sympy.Rational(2,3)*(y-2) \
              + sympy.Rational(10,27)*(x-1)**2 - sympy.Rational(8,27)*(x-1)*(y-2) + sympy.Rational(5,54)*(y-2)**2
        #sympy.pprint(sympy.expand(epx))
        self.assertTrue(sympy.expand(px) == sympy.expand(epx))

    # Test alternate symbols and give x0 as a list of tuples that makes the points unambiguous
    def test_different_symbols_multi_var(self):
        pi = sympy.symbols('pi')
        theta = sympy.symbols('theta')
        kappa = sympy.symbols('kappa')
        f = sympy.sin(kappa) + sympy.cos(theta) + sympy.sin(3*pi)
        x0 = [2*math.pi/3, math.pi/3, 2*math.pi/3]
        px = taylor(f, list(zip([pi, theta, kappa], x0)), 13)

        xeval = [(pi, math.pi/2), (theta, math.pi/3), (kappa, math.pi/4)]
        #sympy.pprint(f.subs(xeval))
        self.assertAlmostEqual(f.subs(xeval).evalf(), px.subs(xeval).evalf())

    # Test R^N->R^M
    def test_rn_to_rm(self):
        f1 = sympy.functions.elementary.exponential.exp(x+y)
        f2 = sympy.functions.elementary.trigonometric.cos(x**2 + y + z)
        f = (f1, f2)
        x0 = [0, 0]
        x1 = (0,0, 1)
        p = taylor(f, (x0, x1), 2)

        # Check that the components match up
        self.assertEqual(taylor(f1, x0, 2), p[0])
        self.assertEqual(taylor(f1, x0, 2), p[0])
        self.assertEqual(taylor(f2, x1, 2), p[1])

        # Check that the list notation also works
        p = taylor([f1, f2], [x0, x1], 2)
        self.assertEqual(taylor(f1, x0, 2), p[0])
        self.assertEqual(taylor(f2, x1, 2), p[1])

        f3 = sympy.functions.elementary.trigonometric.sin(y)
        x3 = (0,)
        p = taylor((f1, f2, f3), (x0, x1, x3), 2)
        self.assertEqual(taylor(f1, x0, 2), p[0])
        self.assertEqual(taylor(f2, x1, 2), p[1])
        self.assertEqual(taylor(f3, x3, 2), p[2])



import unittest

from testnum import TestNumericalTaylor
from testsym import TestSymbolicTaylor
from testutils import TestTaylorUtils
from testdiffcalc import TestDiffCalc


def main():
    unittest.main()


if __name__ == "__main__":
    main()


from setuptools import setup

setup(
    name='ohtaylor',
    version='1.0.0',
    packages=['ohtaylor'],
    url='https://gitlab.com/mochan/ohtaylor',
    license='Apache 2.0',
    author='Mochan Shrestha',
    author_email='',
    description='Taylor Series calculator'
)




import unittest, sympy

from ohtaylor.diffcalc import *
from sympy.abc import x, y, z

class TestDiffCalc (unittest.TestCase):

    # Test that unzipping works
    def test_unzipping(self):
        vars = zip([x, y, z], [1, 2, 3])
        vars, x0 = zip(*vars)
        self.assertEqual((x, y, z), vars)
        self.assertEqual((1, 2, 3), x0)

    # Check that the basic diff calculating works
    def test_basic_diff(self):
        d = DiffCalc()
        f = x**2 * y**2 * z**2
        df = d.diff(f, (x, y, z), (1, 1, 1))
        #sympy.pprint(df)
        self.assertEqual(8*x*y*z, df)

    # Check that getting the smaller tuple value works
    def test_get_smaller_tuple(self):
        zt = DiffCalcCached.get_smaller_tuple((0,0,0))
        self.assertEqual((-1, (0, 0, 0)), zt)

        t2 = (1, 2, 3)
        t2r = DiffCalcCached.get_smaller_tuple(t2)
        self.assertEqual((0, (0, 2, 3)), t2r)

        t3 = (0, 3, 2)
        t3r = DiffCalcCached.get_smaller_tuple(t3)
        self.assertEqual((1, (0, 2, 2)), t3r)

    # Test the cached diff method
    def test_cached_diff(self):
        d = DiffCalcCached()
        f = x**2 * y**2 * z**2
        symbs = (x, y, z)
        f1 = d.diff(f, symbs, (0, 0, 0))
        self.assertEqual(f, f1)
        f1 = d.diff(f, symbs, (1, 0, 0))
        self.assertEqual(2*x * y**2 * z**2, f1)
        t = (2, 0, 0)
        i1, t1 = DiffCalcCached.get_smaller_tuple(t)
        self.assertTrue(t1 in d.diff_cache)
        self.assertEqual(0, i1)
        f1 = d.diff(f, symbs, (2, 0, 0))
        self.assertEqual(2 * y ** 2 * z ** 2, f1)

        f1 = d.diff(f, symbs, (0, 0, 2))
        self.assertEqual(2 * y ** 2 * x ** 2, f1)





import sympy


# Calculate the derivatives using diff and indices in the tuple
class DiffCalc:
    # Take the derivative of the function f according to the tuple t with the order given in x
    @staticmethod
    def diff(f, x, t):
        k = len(x)
        fix = f

        for j in range(0, k):
            fix = sympy.diff(fix, x[j], t[j])

        return fix


# Calculate the derivatives but try to cache previous derivatives
class DiffCalcCached:
    # Cached values of previous diffs from tuples of lower sum
    diff_cache = {}

    # Find a tuple that is one smaller in sum. It will go from left to right
    # to reduce the sum. If a value is 0, it will move to the right one step
    @staticmethod
    def get_smaller_tuple(t):
        val = sum(t)

        # If the sum is zero, then we cannot find a smaller tuple
        if val == 0:
            return -1, t

        for i in range(0, len(t)):
            if t[i] > 0:
                return i, t[:i] + (t[i] - 1,) + t[i+1:]

    def diff(self, f, x, t):
        # Nothing to take derivative of. Add f to the diff cache
        if sum(t) == 0:
            self.diff_cache[t] = f
            return f

        # Look up the cache to find a previously calculated derivative that we
        # just have to do one more diff of to get the derivative we want
        # If it is not found, just calculate the entire derivative and add it to cache
        i, st = DiffCalcCached.get_smaller_tuple(t)
        if st in self.diff_cache:
            dfi = self.diff_cache[st]
            df = sympy.diff(dfi, x[i])
        else:
            df = DiffCalc.diff(f, x, t)

        # Add the calculated derivative to the cache
        self.diff_cache[t] = df
        return df